#!/System/Library/Frameworks/Ruby.framework/Versions/1.8/usr/bin/ruby

#  bundle_check-appstore.rb
#  
#  Copyright (c) 2011 Realmac Software. All rights reserved.

require 'osx/cocoa'
include OSX

$libraryDirectoryPath = File.dirname(__FILE__)
$LOAD_PATH << $libraryDirectoryPath

require 'codesign.rb'
require 'bundle.rb'

def usage
	print "Usage:\n"
	print "-applicationPath: path to the app bundle to be verified and archived [required]\n"
	print "-fixturePath: path to the fixture document use for validation [required]\n"
	print "\n"
	print "Example:\n"
	print "ruby application_check.rb\\\n -applicationPath ~/Projects/Products/Release\\ -\\ Standard/Courier.app\\\n -fixturePath ~/Projects/Source/Courier/scripts/courier-standard.plist\n"
	print "\n"
	exit 1
end

if ARGV.count == 0
	usage
end

def start
	print "============  BUNDLE VALIDATION STARTED ============\n"
	print "\n"
end

def _printFail(reason, severity)
	print "\n"
	print "============  BUNDLE VALIDATION FAILURE ============\n"
	
	if ! reason.nil? and ! severity.nil?
		print "#{severity.to_s}: #{reason}\n"
		print "\n"
	end
end

def fail(reason = nil, severity = :warning)
	_printFail(reason, severity)
	
	orderOfSeverity = [
		:advisory,
		:warning,
		:fatal
	]
	if ! $severity.nil? and orderOfSeverity.index(severity) <= orderOfSeverity.index($severity)
		return
	end
	
	if severity == :fatal
		exit 1
	elsif severity == :warning
		def finish
			_printFail "warnings were issued, see notices above", :fatal
			exit 1
		end
	elsif severity == :advisory
		def finish
			_printFail "warnings were issued, see notices above", :advisory
			exit 0
		end
	end
	
	$severity = severity
end

def finish
	print "\n"
	print "============  BUNDLE VALIDATION SUCCESS ============\n"
end


# Check arguments

mainBundlePath = NSUserDefaults.standardUserDefaults.stringForKey "applicationPath"
if mainBundlePath.nil?
	fail "an application path must be specified using -applicationPath", :fatal
end
unless NSFileManager.defaultManager.fileExistsAtPath mainBundlePath
	fail "bundle doesn't exist at the specified path", :fatal
end
$mainBundleURL = NSURL.fileURLWithPath mainBundlePath
mainBundle = NSBundle.bundleWithURL $mainBundleURL
if mainBundle.nil?
	fail "couldn't load application bundle", :fatal
end

fixturePath = NSUserDefaults.standardUserDefaults.stringForKey "fixturePath"
if fixturePath.nil?
	fail "a fixture path must be specified using -fixturePath", :fatal
end
unless NSFileManager.defaultManager.fileExistsAtPath fixturePath
	fail "fixture doesn't exist at the specified path", :fatal
end
$fixtureURL = NSURL.fileURLWithPath fixturePath


# Utility functions

def _lookupFixtureValuesForKey(fixtureLocation = nil, key = "", depth = :shallow)
	# \param depth
	# :shallow - return the first match
	# :deep - return an array of all matches ordered from deepest to shallowest
	# :flatten - compose the higher values on top of the lower values, objects must all be of the same class for this to work
	
	if depth == :shallow
		allValues = _lookupFixtureValuesForKey fixtureLocation, key, :deep
		return allValues.lastObject
	end
	
	if depth == :flatten
		allValues = _lookupFixtureValuesForKey fixtureLocation, key, :deep
		
		compositeValue = nil
		append = nil
		
		lastObject = allValues.lastObject
		if lastObject.nil?
			# nop
		elsif lastObject.isKindOfClass NSArray.objc_send(:class)
			compositeValue = NSMutableArray.array
			
			append = lambda do |val|
				compositeValue.addObjectsFromArray val
			end
		elsif lastObject.isKindOfClass NSDictionary.objc_send(:class)
			compositeValue = NSMutableDictionary.dictionary
			
			append = lambda do |val|
				compositeValue.addEntriesFromDictionary val
			end
		end
		
		allValues.each do |currentValue|
			append.call currentValue
		end
		
		return compositeValue
	end
	
	if depth != :deep
		throw "unknown enumeration type #{depth}"
		exit 1
	end
	
	values = NSMutableArray.array
	
	fixture = NSDictionary.dictionaryWithContentsOfURL fixtureLocation
	
	fixtureImport = fixture.objectForKey "Import"
	unless fixtureImport.nil?
		currentFixtureLocation = fixtureLocation.URLByDeletingLastPathComponent.URLByAppendingPathComponent fixtureImport
		values.concat _lookupFixtureValuesForKey(currentFixtureLocation, key, :deep)
	end
	
	currentValue = fixture.objectForKey key
	unless currentValue.nil?
		values << currentValue
	end
	
	values
end

def lookupFixtureValuesForKey(key = "", depth = :shallow)
	_lookupFixtureValuesForKey $fixtureURL, key, depth
end


#

start


# Bundle Checks

print "Checking #{mainBundlePath} for validity...\n"
print "\n"


# Check for version control artefacts

print "Checking for version control artefacts...\n"

def recurseIntoFolderLookingForVersionControlArtefacts(folderLocation = nil)
	versionControlArtefactFolderNames = NSArray.arrayWithObjects ".svn", ".git", nil
	
	fileDetails = NSFileManager.defaultManager.fileExistsAtPath_isDirectory folderLocation.path
	return unless fileDetails[0] && fileDetails[1]
	
	folderPath = folderLocation.path
	folderEnumerator = NSFileManager.defaultManager.enumeratorAtPath folderPath
	
	artefactsFound = false
	
	begin
		currentPath = folderEnumerator.nextObject
		break if currentPath.nil?
		
		currentFullPath = folderPath.stringByAppendingPathComponent currentPath
		
		versionControlArtefactFolderNames.each do |currentArtefactName|
			next unless currentFullPath.lastPathComponent.isEqualToString currentArtefactName
			
			print "version control artefact found at #{currentFullPath}\n"
			artefactsFound = true
			
			folderEnumerator.skipDescendants
		end
	end while true
	
	fail "version control artefacts found" if artefactsFound
end
recurseIntoFolderLookingForVersionControlArtefacts(mainBundle.bundleURL)
print "\n"

# Check for _MASReceipt folder

print "Checking bundle doesn't contain a _MASReceipt folder...\n"

fail "application bundle contains a _MASReceipt folder" if NSFileManager.defaultManager.fileExistsAtPath $mainBundleURL.URLByAppendingPathComponent("Contents").URLByAppendingPathComponent("_MASReceipt").path


# Bundle Check Functions

def architectureValuesForNames(architectureNames)
	architectureNameToBundleValueMap = {
		"i386" => NSBundleExecutableArchitectureI386,
		"ppc" => NSBundleExecutableArchitecturePPC,
		"x86_64" => NSBundleExecutableArchitectureX86_64,
		"ppc64" => NSBundleExecutableArchitecturePPC64,
	}
	
	architectureValues = NSMutableArray.array
	
	architectureNames.each do |currentName|
		currentValue = architectureNameToBundleValueMap[currentName.to_ruby]
		fail "unknown value for architecture name #{currentName}", :fatal if currentValue.nil?
		
		architectureValues << currentValue
	end
	
	architectureValues
end

def checkBundleExecutableArchitecturesContainsAllOf(location = nil, presentArchitectures = nil, expectedArchitectures = nil)
	return if location.nil?
	return if presentArchitectures.nil?
	
	presentExpectedArchitectures = expectedArchitectures.select do |currentArchitecture|
		presentArchitectures.include? currentArchitecture
	end
	fail "The executable at #{location.path} doesn't contain all the expected architectures, architectures found #{presentArchitectures}, expected #{expectedArchitectures}." unless presentExpectedArchitectures.length == expectedArchitectures.length
end

def _shouldCheck(checkFixture, bundleIdentifier)
	if checkFixture.nil?
		return false
	end
	
	if checkFixture.isKindOfClass NSNumber.objc_send(:class)
		return checkFixture.isEqual true.to_ns
	end
	
	if checkFixture.isKindOfClass NSDictionary.objc_send(:class)
		excludeBundles = checkFixture.objectForKey "exclude"
		return !excludeBundles.containsObject(bundleIdentifier)
	end
	
	raise "unknown check fixture class"
end

def bundleCheck(bundleLocation = nil)
	print "Checking #{bundleLocation.lastPathComponent}...\n"
	print "\n"
	
	
	currentBundleExecutableArchitectures = CFBundleCopyExecutableArchitecturesForURL bundleLocation
	bundleCodeSignRequirements = lookupFixtureValuesForKey "BundleCodeSign", :flatten
	unless currentBundleExecutableArchitectures.nil? || bundleCodeSignRequirements.nil?
		print "> Checking codesign validity...\n"
		
		compositeRequirementString = bundleCodeSignRequirements.componentsJoinedByString " and "
		puts compositeRequirementString
		
		codesignValid, codesignValidError = Codesign.validateBundleAtLocationWithRequirementReturningError bundleLocation, compositeRequirementString
		fail "#{bundleLocation.path} codesign signature not valid, #{codesignValidError.localizedDescription}" unless codesignValid
	end
	
	
	bundleCheckDescriptionToScriptNameMap = {
		"localisations" => "bundle_checklocalisations.rb",
		"images" => "bundle_checkimages.rb",
	}
	
	bundleIdentifier = NSBundle.bundleWithURL(bundleLocation).bundleIdentifier
	
	bundleChecks = []
	bundleChecks << "localisations" if _shouldCheck(lookupFixtureValuesForKey("BundleCheckLocalisations"), bundleIdentifier)
	bundleChecks << "images" if _shouldCheck(lookupFixtureValuesForKey("BundleCheckImages"), bundleIdentifier)
	
	bundleChecks.each do |description|
		print " > Checking #{description}...\n"
		scriptName = bundleCheckDescriptionToScriptNameMap[description]
		
		$stdout.flush
		checkTask = NSTask.launchedTaskWithLaunchPath_arguments_ $libraryDirectoryPath.to_ns.stringByAppendingPathComponent(scriptName), [ "-bundlePath", bundleLocation.path ].to_ns
		checkTask.waitUntilExit
		
		unless checkTask.terminationStatus == 0
			fail " > #{bundleLocation.lastPathComponent} failed #{description} validation\n"
			next
		end
	end
	
	print "\n"
end


# Main Bundle Checks

infoPlistChecks = lookupFixtureValuesForKey "ApplicationInfoPlist", :flatten
unless infoPlistChecks.nil?
	print "Checking main bundle Info.plist...\n"
	
	infoPlistChecks.each do |key, value|
		expectedValue = value
		bundleValue = mainBundle.infoDictionary.objectForKey key
		
		if expectedValue.isEqual bundleValue
			next
		end
		
		# Try type coercion
		
		if expectedValue.isKindOfClass NSNumber.objc_send(:class)
			shouldBePresent = expectedValue.isEqual true.to_ns
			isPresent = ! bundleValue.nil?
			
			next if shouldBePresent == isPresent
		end
		
		fail "bundle Info.plist value for #{key} doesn't match the expected value of #{expectedValue}"
	end
end

applicationArchitectureNames = lookupFixtureValuesForKey "ApplicationArchitectures", :flatten
unless applicationArchitectureNames.nil?
	print "Checking main bundle executable architectures...\n"
	
	mainBundleExpectedArchitectures = architectureValuesForNames applicationArchitectureNames
	
	checkBundleExecutableArchitecturesContainsAllOf mainBundle.bundleURL, mainBundle.executableArchitectures, mainBundleExpectedArchitectures
	unexpectedArchitecturesPresent = mainBundle.executableArchitectures.reject do |currentArchitecture|
		mainBundleExpectedArchitectures.include? currentArchitecture
	end
	fail "The bundle #{mainBundle.executablePath} contains architectures that weren't expected, architectures found #{mainBundle.executableArchitectures}, expected #{mainBundleExpectedArchitectures}" unless unexpectedArchitecturesPresent.length == 0
end

print "Checking main bundle has all the required components...\n"
iconResourceName = mainBundle.objectForInfoDictionaryKey "CFBundleIconFile"
if iconResourceName.nil?
	fail "bundle doesn't have an icon listed CFBundleIconFile"
else
	iconLocation = mainBundle.URLForImageResource iconResourceName
	
	getTypeAndError = iconLocation.getResourceValue_forKey_error NSURLTypeIdentifierKey
	if ! getTypeAndError[0]
		fail "couldn't find bundle image resource for name #{iconResourceName}"
	elsif NSWorkspace.sharedWorkspace.type_conformsToType_ getTypeAndError[1], "com.apple.icns"
		taskPath = $libraryDirectoryPath.to_ns.stringByAppendingPathComponent_ "icns_check.rb"
		
		taskArguments = [ "-icnsPath", iconLocation.path ]
		
		minimumOperatingSystem = mainBundle.infoDictionary.objectForKey "LSMinimumSystemVersion"
		unless minimumOperatingSystem.nil?
			minimumOperatingSystemComponents = minimumOperatingSystem.componentsSeparatedByString_ "."
			if minimumOperatingSystemComponents.count > 2
				minimumOperatingSystemComponents = minimumOperatingSystemComponents.subarrayWithRange NSMakeRange 0, 2
				minimumOperatingSystem = minimumOperatingSystemComponents.componentsJoinedByString_ "."
			end
			
			taskArguments << "-minimumOperatingSystem"
			taskArguments << minimumOperatingSystem
		end
		
		$stdout.flush
		task = NSTask.launchedTaskWithLaunchPath_arguments_ taskPath, taskArguments
		
		task.waitUntilExit
		
		fail "bundle icon doesn't include all the required components", :advisory unless task.terminationStatus == 0
	end
end

applicationCodeSignRequirements = lookupFixtureValuesForKey "ApplicationCodeSign", :flatten
unless applicationCodeSignRequirements.nil?
	print "Checking main bundle codesign validity...\n"
	
	compositeRequirementString = applicationCodeSignRequirements.componentsJoinedByString " and "
	puts "#{compositeRequirementString}\n"
	
	codesignValid, codesignValidError = Codesign.validateBundleAtLocationWithRequirementReturningError mainBundle.bundleURL, compositeRequirementString
	print "\n"
	
	fail "application bundle is not codesign valid, #{codesignValidError.localizedDescription}" unless codesignValid
end


# Application Bundle Checks

allBundleLocations = Bundle.childBundleLocationsOfBundleAtLocation $mainBundleURL

# There may be duplicate bundle identifiers, ensure the map is of { location => bundle_identifier }
allBundleLocationIdentifierPairs = []
allBundleLocations.each do |currentBundleLocation|
	currentInfoDictionary = CFBundleCopyInfoDictionaryForURL currentBundleLocation
	
	if currentInfoDictionary.nil?
		fail "bundle at location #{currentBundleLocation.absoluteString} doesn't have an Info.plist", :warning
		next
	end
	
	currentBundleIdentifier = currentInfoDictionary.objectForKey_ "CFBundleIdentifier"
	
	if currentBundleIdentifier.nil?
		fail "bundle at location #{currentBundleLocation.absoluteString} doesn't have a bundle identifier", :warning
		next
	end
	
	allBundleLocationIdentifierPairs << [ currentBundleLocation, currentBundleIdentifier ]
end

allBundleLocationToIdentifierMap = Hash[*(allBundleLocationIdentifierPairs.flatten)]

expectedBundleIdentifiers = lookupFixtureValuesForKey "ApplicationBundles", :flatten
unless expectedBundleIdentifiers.nil?
	print "Checking main application child bundles are present...\n"
	
	bundlesPresent = NSSet.setWithArray_(allBundleLocationToIdentifierMap.values.to_ns)
	bundlesExpected = NSSet.setWithArray_(expectedBundleIdentifiers)
	
	allPresent = true
	
	# Unexpected bundles = bundles present - bundles expected
	unexpectedBundleIdentifiers = NSMutableSet.setWithSet_(bundlesPresent)
	unexpectedBundleIdentifiers.minusSet_(bundlesExpected)
	unless unexpectedBundleIdentifiers.count == 0
		fail "application contains unexpected bundles: #{unexpectedBundleIdentifiers}!\n", :warning
		allPresent = false
	end
	
	# Absent bundles = bundles expected - bundles present
	absentBundles = NSMutableSet.setWithSet_(bundlesExpected)
	absentBundles.minusSet_(bundlesPresent)
	unless absentBundles.count == 0
		fail "application is missing expected bundles: #{absentBundles}!\n", :warning
		allPresent = false
	end
	
	print "> all main application child bundles are present\n\n" if allPresent
end


# Apply general bundle checks to the application too

bundleCheck($mainBundleURL)


# Ensure sub bundle bundle identifiers are unique

print "Checking main bundle hierarchy for duplicate bundle identifiers...\n"
bundleLocationToInfoPlistDictionary = NSMutableDictionary.dictionary
allBundleLocations.each do |currentBundleLocation|
	currentBundleLocationInfoPlist = CFBundleCopyInfoDictionaryForURL(currentBundleLocation)
	if currentBundleLocationInfoPlist.nil?
		fail "couldn't read Info.plist for bundle at location #{currentBundleLocation}"
		next
	end
	
	bundleLocationToInfoPlistDictionary.setObject_forKey_ currentBundleLocationInfoPlist, currentBundleLocation
end

bundleLocationToBundleIdentifier = NSMutableDictionary.dictionary
bundleLocationToInfoPlistDictionary.allKeys.each do |currentBundleLocation|
	currentBundleInfoPlist = bundleLocationToInfoPlistDictionary.objectForKey_ currentBundleLocation
	bundleLocationToBundleIdentifier.setValue_forKey_ currentBundleInfoPlist.objectForKey_("CFBundleIdentifier"), currentBundleLocation
end

uniqueBundleIdentifiers = NSSet.setWithArray bundleLocationToBundleIdentifier.allValues
uniqueBundleIdentifiers.allObjects.each do |currentBundleIdentifier|
	bundleLocations = bundleLocationToBundleIdentifier.allKeysForObject_ currentBundleIdentifier
	if bundleLocations.count == 1
		next
	end
	
	bundleLocationsString = bundleLocations.componentsJoinedByString_ "\n"
	
	fail "multiple bundles share the same bundle identifier \"#{currentBundleIdentifier}\"\n#{bundleLocationsString}"
end

print "\n"


# Child Bundle Checks

allBundleLocations.each do |currentBundleLocation|
	currentBundleName = currentBundleLocation.lastPathComponent
	
	currentBundle = NSBundle.bundleWithURL currentBundleLocation
	
	# Check that the bundle contains all the architectures of the main bundle
	
	if currentBundle.nil? || !currentBundle.executablePath.nil?
		print "Checking #{currentBundleName} executable architectures...\n"
		currentBundleLocationExecutableArchitectures = CFBundleCopyExecutableArchitecturesForURL(currentBundleLocation)
		if currentBundleLocationExecutableArchitectures.nil?
			fail "couldn't check bundle executable architectures for bundle at location #{currentBundleLocation}"
		else
			checkBundleExecutableArchitecturesContainsAllOf currentBundleLocation, currentBundleLocationExecutableArchitectures, mainBundle.executableArchitectures
		end
	end
	
	# Apply generic bundle checks
	
	unless currentBundle.nil?
		bundleCheck(currentBundle.bundleURL)
		print "\n"
	end
	
	print "\n"
end


# updates feed check, this MUST come after the Info.plist checks so that the feed has been validated

feedURLString = mainBundle.objectForInfoDictionaryKey "DSUpdatesFeedURI"
unless feedURLString.nil?
	print "Checking updates feed...\n"
	feedURL = NSURL.URLWithString feedURLString
	feedData = NSData.dataWithContentsOfURL feedURL
	feedDocument = NSXMLDocument.alloc.initWithData_options_error feedData, 0
	fail "updates feed data cannot be parsed" unless feedDocument[0]
	feedDocument = feedDocument[0]
	
	print "Checking updates feed versions...\n"
	allVersions = feedDocument.nodesForXPath_error "//@sparkle:version"
	fail "cannot parse updates feed" unless allVersions[0]
	allVersions = allVersions[0]
	
	allVersions = allVersions.valueForKey "stringValue"
	bundleVersion = mainBundle.objectForInfoDictionaryKey "CFBundleVersion"
	allVersions.each do |currentVersion|
		if bundleVersion.compare_options(currentVersion, NSNumericSearch) == NSOrderedDescending
			next
		end
		
		fail "the updates feed contains a version with greater or equal bundle version"
		break
	end
end


#

finish
